import numpy as np
from scipy.signal import butter, filtfilt, welch, tukey, decimate
from gwpy.timeseries import TimeSeries
import matplotlib.mlab as mlab, sys, warnings
from scipy.interpolate import interp1d

def fetch_data(ifo, tstart, tend, channel=None, path=None, verbose=0, tag=None):

    """
        Fetch data for a particular event

        ifo: IFO name ('H1' etc)
        tstart, tend: start and end time to find
        path: Local path to save file. If file exists it will be read from disk rather than fetched
        channel: Channel to read from frame data. If 'GWOSC' will fetch open data
        verbose: Print some info

        Returns a gwpy.TimeSeries object
    """

    # If file was previously saved, open it.
    if path is not None and os.path.exists(path):
        tseries = TimeSeries.read(path,start=tstart,end=tend)
        if verbose:
            sys.stdout.write('Reading from local file '+path)
        return tseries

    # If not, then see if it is on GWOSC.
    if channel=='GWOSC':
        #When downloading public data, fetch them with the highest possible sampling rate, then pyRing will down-sample internally, if required. This is needed to avoid incompatibilities between GWOSC down-sampling and the pyRing internal one. The actual function used to down-sample is the same, but differences in things like the length of data stretch can affect filtering at the borders and hence the Bayes Factors. 
        tseries = TimeSeries.fetch_open_data(ifo, tstart, tend, sample_rate = 16384, verbose=verbose, cache=True, tag=u"{0}".format(tag))
    else:
        # Read from authenticated data.
        if channel is None:
            raise Exception('Channel not specified when fetching frame data.')
        tseries = TimeSeries.get(channel, tstart, tend, verbose=verbose)
    if path is not None:
        tseries.write(path)

    return tseries

def chunks(times, strain, chunksize, avoid=None, window=False, alpha=0.1):
    """
        Divide the data in chunks. Skip the 0th and the last chunks which have filter ringing from downsampling.
    """
    if avoid is None:
        avoid = times[0]-1e6 # dummy value
    if window:
        win      = tukey(chunksize,alpha=alpha)
    else:
        win      = np.ones(chunksize)
    #The integer division is needed in case the chunk length in seconds is not a sub-multiple of the total strain 
    #length (e.g. after quality veto cuts)
    for j in range(1,len(strain)//chunksize):
        if not times[chunksize*j] < avoid < times[chunksize*(j+1)]:
            yield strain[chunksize*j:chunksize*(j+1)]*win
            
def local_load_data(ifo, **kwargs):

    sys.stdout.write('\nReading data...\n')
    # General parameters that are always needed.
    triggertime       = kwargs['trigtime']
    fft_acf           = kwargs['fft-acf']
    signal_chunk_size = kwargs['signal-chunksize']
    noise_chunk_size  = kwargs['noise-chunksize']
    srate             = kwargs['sampling-rate']
    f_min_bp          = kwargs['f-min-bp']
    f_max_bp          = kwargs['f-max-bp']
    alpha_window      = kwargs['alpha-window']
    signal_seglen     = int(srate*signal_chunk_size)
    noise_seglen      = int(srate*noise_chunk_size)
    psd_window        = tukey(noise_seglen,alpha_window)
    psd_window_norm   = np.sum(psd_window**2)/noise_seglen
    assert not(triggertime is None), "No triggertime given."

    # If no data file was passed either download them or generate them yourself.
    if(kwargs['download-data'] == 1):
        T         = float(kwargs['datalen-download'])
        channel   = kwargs['channel-{}'.format(ifo)]
        starttime = int(triggertime)-(T/2.)
        endtime   = int(triggertime)+(T/2.)
        #starttime = int(triggertime - T/2.)+1
        #endtime   = int(triggertime + T/2.)+1
        sys.stdout.write('\nUsing GWPY to download data.\n')
        tag        = kwargs['tag']
        tseries    = fetch_data(ifo, starttime, endtime, channel=channel, path=None, verbose=2, tag=tag)
        
        rawtime                 = np.array(tseries.times)
        rawstrain               = np.array(tseries.data)
        
        dt                      = tseries.dt.to_value()
        srate_dt                = tseries.sample_rate.to_value()
        sys.stdout.write('\nLoaded channel {} starting at {} length {}s.\n'.format(channel,starttime,T))
    else:
        raise Exception("Either pass a data file, download data or generate data.")

    assert not((noise_chunk_size > T) or (signal_chunk_size > T)), \
    "Noise ({} s) and signal ({} s) seglens must be shorter than data duration ({})".format(
        noise_chunk_size, signal_chunk_size, T)

    if (not(kwargs['data-{0}'.format(ifo)]=='') or (kwargs['download-data'] == 1)):
        # If there are nans, resize the strain to the longest segment which contains no nans.
        no_nans = 0
        while(no_nans == 0):
            if((np.isnan(rawstrain).any())):
                sys.stdout.write('Nans found in the data. Resizing strain.\n')
                new_rawstrain, new_starttime = resize_nan_strain(
                    rawstrain, starttime, triggertime, signal_chunk_size, dt)
                starttime = new_starttime
                rawstrain = new_rawstrain
            else:
                no_nans = 1
        # Recompute the total length after resizing.
        T = len(rawstrain)*dt

    if(kwargs['bandpassing']):
        # Bandpassing section.
        sys.stdout.write('Bandpassing the raw strain between [{}, {}] Hz.\n'.format(f_min_bp, f_max_bp))
        # Create a Butterworth bandpass filter between [f_min, f_max] and apply it with the function filtfilt.
        bb, ab = butter(4, [f_min_bp/(0.5*srate_dt), f_max_bp/(0.5*srate_dt)], btype='band')
        strain = filtfilt(bb, ab, rawstrain)
    else:
        sys.stdout.write('No bandpassing applied.\n')
        strain = rawstrain
    # Check that the sample rate from the data is the same passed in the configuration file. 
    # In case they are different, either downsample or throw an error.
    if (srate > srate_dt):
        raise ValueError("You requested a sample rate higher than the data sampling.")
    elif (srate < srate_dt):
        sys.stdout.write('Downsampling detector data from {} to {} Hz, decimate factor {}\n'.format(
            srate_dt, srate, int(srate_dt/srate)))
        strain = decimate(strain, int(srate_dt/srate), zero_phase=True)
        dt     = 1./srate
    else:
        pass
    
    bandpasstime = rawtime[::int(srate_dt/srate)]

    # Compute the index of the trigtime 
    # (estimate of the coalescence time of a signal or the time at which the injection should be placed).
    index_trigtime = int((triggertime-starttime)*srate)

    # Find the on-source chunk, centred on triggertime.
    mask = np.ones(len(strain), dtype=bool)
    if not((signal_seglen%2)==0):
        on_source_strain = strain[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2+1]
        on_source_time   = bandpasstime[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
        mask[range(index_trigtime-signal_seglen//2,index_trigtime+signal_seglen//2+1,1)] = False
    else:
        on_source_strain = strain[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
        on_source_time   = bandpasstime[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
        mask[range(index_trigtime-signal_seglen//2,index_trigtime+signal_seglen//2,1)]   = False

    #Excise the on-source chunk to avoid including the signal in the PSD computation.
    psd_strain = tuple((strain)[mask])

    # While it is true that no FFT is applied to the on-source chunk, the ACF is computed on a windowed chunk, \
    #thus to make a consistent likelihood computation, also the onsource chunk must be windowed when no truncation 
    #is applied.
    if (kwargs['window-onsource'] and kwargs['window']):
        if(kwargs['truncate']):
            warnings.warn("The on-source chunk should not be windowed when truncating data.")
        else:
            assert (signal_seglen==noise_seglen), \
            "If a window is applied, the length of the signal chunk and of the noise chunk must be the same, \
            otherwise with the same Tukey alpha-parameter PSD will be underestimated. Either choose different \
            alphas or equal lengths."
        on_source_window = tukey(signal_seglen, alpha=alpha_window)
        on_source_strain = on_source_strain*on_source_window

    #times = np.linspace(starttime, starttime+T-dt, len(strain))

    if not(kwargs['signal-chunksize']==kwargs['noise-chunksize']):
        warnings.warn("A different chunksize between signal and noise implies an incorrect normalization \
        of the autocorrelation function. This configuration should not be used for production runs.")
    
    del rawstrain, strain

    return on_source_strain, on_source_time

def pyring_time(model,local_wheel_time):
    waveform_time_array = {}
    dt = {}
    t_start = 0
    for det in ['H1','L1']:
        dt[det]   = model.time_delay['H1_'+det]
        #time_array_raw = model.detectors[det].time - (model.tevent+dt[det])
        time_array_raw = local_wheel_time[det] - (model.tevent + dt[det])
        waveform_time_array[det] = time_array_raw[time_array_raw >= t_start][:model.duration_n] 
    return waveform_time_array,dt